from multiprocessing import cpu_count

import click
from loguru import logger

from reprobench.console.decorators import server_info, common, use_tunneling

from .manager import LocalManager


@click.command("local")
@click.option("-w", "--num-workers", type=int, default=cpu_count(), show_default=True)
@click.option(
    "-d", "--output-dir", type=click.Path(), default="./output", show_default=True
)
@click.option("-r", "--repeat", type=int, default=1)
@click.argument("command", type=click.Choice(("run",)))
@click.argument("config", type=click.Path(), default="./benchmark.yml")
@server_info
@use_tunneling
@common
def cli(command, **kwargs):
    manager = LocalManager(**kwargs)

    if command == "run":
        manager.run()

    # TODO: add run_with_server


if __name__ == "__main__":
    cli()



from pathlib import Path

import click
import gevent
import zmq.green as zmq
from loguru import logger
from reprobench.console.decorators import common, server_info
from reprobench.core.bootstrap.server import bootstrap
from reprobench.core.db import Observer
from reprobench.core.events import BOOTSTRAP
from reprobench.core.observers import CoreObserver
from reprobench.utils import decode_message, import_class


class BenchmarkServer(object):
    BACKEND_ADDRESS = "inproc://backend"

    def __init__(self, frontend_address, **kwargs):
        self.frontend_address = frontend_address

    def receive_event(self):
        address, event_type, payload = self.frontend.recv_multipart()
        logger.trace((address, event_type, decode_message(payload)))
        return address, event_type, payload

    def loop(self):
        while True:
            address, event_type, payload = self.receive_event()
            self.backend.send_multipart([event_type, payload, address])

    def run(self):
        self.context = zmq.Context()
        self.frontend = self.context.socket(zmq.ROUTER)
        self.frontend.bind(self.frontend_address)
        self.backend = self.context.socket(zmq.PUB)
        self.backend.bind(self.BACKEND_ADDRESS)

        core_observer_greenlet = gevent.spawn(
            CoreObserver.observe,
            self.context,
            backend_address=self.BACKEND_ADDRESS,
            reply=self.frontend,
        )
        logger.info(f"Listening on {self.frontend_address}...")

        serverlet = gevent.spawn(self.loop)
        logger.info(f"Ready to receive events...")
        serverlet.join()
        core_observer_greenlet.kill()


@click.command(name="server")
@server_info
@common
def cli(server_address, **kwargs):
    server = BenchmarkServer(server_address, **kwargs)
    server.run()


if __name__ == "__main__":
    cli()



import atexit
import itertools
import json
import shutil
from pathlib import Path

import gevent
from loguru import logger
from peewee import chunked
from tqdm import tqdm

from reprobench.core.db import (
    MODELS,
    Limit,
    Observer,
    Parameter,
    ParameterGroup,
    Run,
    Step,
    Task,
    TaskGroup,
    Tool,
    db,
)
from reprobench.utils import (
    check_valid_config_space,
    get_db_path,
    import_class,
    init_db,
    is_range_str,
    parse_pcs_parameters,
    str_to_range,
)

try:
    from ConfigSpace.read_and_write import pcs
except ImportError:
    pcs = None


def bootstrap_db(output_dir):
    db_path = get_db_path(output_dir)
    init_db(db_path)
    db.connect()
    db.create_tables(MODELS, safe=True)


def bootstrap_limits(config):
    # TODO: handle limit changes
    query = Limit.insert_many(
        [{"key": key, "value": value} for (key, value) in config["limits"].items()]
    ).on_conflict("ignore")
    query.execute()


def bootstrap_steps(config):
    count = Step.select().count()
    new_steps = config["steps"]["run"][count:]
    if len(new_steps) > 0:
        query = Step.insert_many(
            [
                {
                    "category": "run",
                    "module": step["module"],
                    "config": json.dumps(step.get("config", None)),
                }
                for step in new_steps
            ]
        )
        query.execute()


def bootstrap_observers(config, observe_args):
    count = Observer.select().count()
    new_observers = config["observers"][count:]
    if len(new_observers) > 0:
        query = Observer.insert_many(
            [
                {
                    "module": observer["module"],
                    "config": json.dumps(observer.get("config", None)),
                }
                for observer in new_observers
            ]
        )
        query.execute()

        for observer in new_observers:
            observer_class = import_class(observer["module"])
            gevent.spawn(observer_class.observe, *observe_args)


def register_steps(config):
    logger.info("Registering steps...")
    for step in itertools.chain.from_iterable(config["steps"].values()):
        import_class(step["module"]).register(step.get("config", {}))


def bootstrap_tasks(config):
    for (name, tasks) in config["tasks"].items():
        TaskGroup.insert(name=name).on_conflict("ignore").execute()
        with db.atomic():
            for batch in chunked(tasks, 100):
                query = Task.insert_many(
                    [{"path": task, "group": name} for task in batch]
                ).on_conflict("ignore")
                query.execute()


def create_parameter_group(tool, group, parameters):
    PCS_KEY = "__pcs"
    pcs_parameters = {}
    use_pcs = PCS_KEY in parameters
    config_space = None

    if use_pcs:
        pcs_text = parameters.pop(PCS_KEY)
        lines = pcs_text.split("\n")
        config_space = pcs.read(lines)
        pcs_parameters = parse_pcs_parameters(lines)

    ranged_enum_parameters = {
        key: value
        for key, value in parameters.items()
        if isinstance(parameters[key], list)
    }

    ranged_numbers_parameters = {
        key: str_to_range(value)
        for key, value in parameters.items()
        if isinstance(value, str) and is_range_str(value)
    }

    ranged_parameters = {
        **pcs_parameters,
        **ranged_enum_parameters,
        **ranged_numbers_parameters,
    }

    if len(ranged_parameters) == 0:
        parameter_group, _ = ParameterGroup.get_or_create(name=group, tool=tool)

        for (key, value) in parameters.items():
            query = Parameter.insert(
                group=parameter_group, key=key, value=value
            ).on_conflict("replace")
            query.execute()
        return

    constant_parameters = {
        key: value for key, value in parameters.items() if key not in ranged_parameters
    }

    tuples = [
        [(key, value) for value in values] for key, values in ranged_parameters.items()
    ]

    for combination in itertools.product(*tuples):
        parameters = {**dict(combination), **constant_parameters}

        if use_pcs:
            check_valid_config_space(config_space, parameters)

        combination_str = ",".join(f"{key}={value}" for key, value in combination)
        group_name = f"{group}[{combination_str}]"

        parameter_group, _ = ParameterGroup.get_or_create(name=group_name, tool=tool)

        for (key, value) in parameters.items():
            query = Parameter.insert(
                group=parameter_group, key=key, value=value
            ).on_conflict("replace")
            query.execute()


def bootstrap_tools(config):
    logger.info("Bootstrapping tools...")

    for tool_name, tool in config["tools"].items():
        query = Tool.insert(name=tool_name, module=tool["module"]).on_conflict(
            "replace"
        )
        query.execute()

        if "parameters" not in tool:
            create_parameter_group(tool_name, "default", {})
            continue

        for group, parameters in tool["parameters"].items():
            create_parameter_group(tool_name, group, parameters)


def bootstrap_runs(config, output_dir, repeat=1):
    parameter_groups = ParameterGroup.select().iterator()
    tasks = Task.select().iterator()
    total = ParameterGroup.select().count() * Task.select().count()

    with db.atomic():
        for (parameter_group, task) in tqdm(
            itertools.product(parameter_groups, tasks),
            desc="Bootstrapping runs",
            total=total,
        ):
            for iteration in range(repeat):
                directory = (
                    Path(output_dir)
                    / parameter_group.tool_id
                    / parameter_group.name
                    / task.group_id
                    / Path(task.path).name
                    / str(iteration)
                )

                query = Run.insert(
                    id=directory,
                    tool=parameter_group.tool_id,
                    task=task,
                    parameter_group=parameter_group,
                    status=Run.PENDING,
                    iteration=iteration,
                ).on_conflict("ignore")
                query.execute()


def bootstrap(config=None, output_dir=None, repeat=1, observe_args=None):
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    bootstrap_db(output_dir)
    bootstrap_limits(config)
    bootstrap_steps(config)
    bootstrap_observers(config, observe_args)
    register_steps(config)
    bootstrap_tasks(config)
    bootstrap_tools(config)
    bootstrap_runs(config, output_dir, repeat)



import math
import subprocess
import sys
from pathlib import Path

from loguru import logger

from sshtunnel import SSHTunnelForwarder
from reprobench.managers.base import BaseManager
from reprobench.utils import read_config

from .utils import to_comma_range


class SlurmManager(BaseManager):
    def prepare(self):
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        limits = self.config["limits"]
        time_limit_minutes = int(math.ceil(limits["time"] / 60.0))

        self.cpu_count = limits.get("cores", 1)
        # @TODO improve this
        self.time_limit = 2 * time_limit_minutes
        self.mem_limit = 2 * limits["memory"]

        if self.tunneling is not None:
            self.server = SSHTunnelForwarder(
                self.tunneling["host"],
                remote_bind_address=("127.0.0.1", self.tunneling["port"]),
                ssh_pkey=self.tunneling["key_file"],
                ssh_config_file=self.tunneling["ssh_config_file"],
            )

            # https://github.com/pahaz/sshtunnel/issues/138
            if sys.version_info[0] > 3 or (
                sys.version_info[0] == 3 and sys.version_info[1] >= 7
            ):
                self.server.daemon_forward_servers = True

            self.server.start()
            self.server_address = f"tcp://127.0.0.1:{self.server.local_bind_port}"
            logger.info(f"Tunneling established at {self.server_address}")

    def stop(self):
        subprocess.run(["scancel", f"--name={self.config['title']}-benchmark-worker"])

    def spawn_workers(self):
        logger.info("Spawning workers...")

        address_args = f"--address={self.server_address}"
        if self.tunneling is not None:
            address_args = f"-h {self.tunneling['host']} -p {self.tunneling['port']} -K {self.tunneling['key_file']}"

        worker_cmd = f"{sys.exec_prefix}/bin/reprobench worker {address_args} -vv"
        worker_submit_cmd = [
            "sbatch",
            "--parsable",
            f"--array=1-{self.pending}",
            f"--time={self.time_limit}",
            f"--mem={self.mem_limit}",
            f"--cpus-per-task={self.cpu_count}",
            f"--job-name={self.config['title']}-benchmark-worker",
            f"--output={self.output_dir}/slurm-worker_%a.out",
            "--wrap",
            f"srun {worker_cmd}",
        ]
        logger.trace(worker_submit_cmd)
        self.worker_job = subprocess.check_output(worker_submit_cmd).decode().strip()
        logger.info(f"Worker job array id: {self.worker_job}")

    def wait(self):
        if self.tunneling is not None:
            self.server.stop()




import sys
import atexit
import json
from pathlib import Path

import click
import zmq
from sshtunnel import SSHTunnelForwarder
from loguru import logger

from reprobench.console.decorators import common, server_info, use_tunneling
from reprobench.core.events import (
    RUN_FINISH,
    RUN_INTERRUPT,
    RUN_START,
    RUN_STEP,
    WORKER_JOIN,
    WORKER_LEAVE,
)
from reprobench.utils import decode_message, import_class, send_event

REQUEST_TIMEOUT = 15000


class BenchmarkWorker:
    def __init__(self, server_address, tunneling=None):
        self.server_address = server_address

        if tunneling is not None:
            self.server = SSHTunnelForwarder(
                tunneling["host"],
                remote_bind_address=("127.0.0.1", tunneling["port"]),
                ssh_pkey=tunneling["key_file"],
                ssh_config_file=tunneling["ssh_config_file"],
            )

            # https://github.com/pahaz/sshtunnel/issues/138
            if sys.version_info[0] > 3 or (
                sys.version_info[0] == 3 and sys.version_info[1] >= 7
            ):
                self.server.daemon_forward_servers = True

            self.server.start()
            self.server_address = f"tcp://127.0.0.1:{self.server.local_bind_port}"
            logger.info(f"Tunneling established at {self.server_address}")
            atexit.register(self.stop_tunneling)

    def killed(self, run_id):
        send_event(self.socket, RUN_INTERRUPT, run_id)
        send_event(self.socket, WORKER_LEAVE)

    def stop_tunneling(self):
        self.server.stop()

    def run(self):
        context = zmq.Context()
        self.socket = context.socket(zmq.DEALER)
        logger.debug(f"Connecting to {self.server_address}")
        self.socket.connect(self.server_address)

        send_event(self.socket, WORKER_JOIN)
        run = decode_message(self.socket.recv())

        self.run_id = run["id"]
        atexit.register(self.killed, self.run_id)

        tool = import_class(run["tool"])

        if not tool.is_ready():
            tool.setup()

        context = {}
        context["socket"] = self.socket
        context["tool"] = tool
        context["run"] = run
        logger.info(f"Processing task: {run['id']}")

        directory = Path(run["id"])
        directory.mkdir(parents=True, exist_ok=True)

        payload = dict(tool_version=tool.version(), run_id=self.run_id)
        send_event(self.socket, RUN_START, payload)

        for runstep in run["steps"]:
            logger.debug(f"Running step {runstep['module']}")
            step = import_class(runstep["module"])
            config = json.loads(runstep["config"])
            step.execute(context, config)
            payload = {"run_id": self.run_id, "step": runstep["module"]}
            send_event(self.socket, RUN_STEP, payload)

        send_event(self.socket, RUN_FINISH, self.run_id)
        atexit.unregister(self.killed)
        send_event(self.socket, WORKER_LEAVE, self.run_id)


@click.command("worker")
@server_info
@use_tunneling
@common
def cli(**kwargs):
    worker = BenchmarkWorker(**kwargs)
    worker.run()


if __name__ == "__main__":
    cli()



from loguru import logger

from reprobench.task_sources.doi import DOISource
from reprobench.task_sources.file import FileSource
from reprobench.task_sources.url import UrlSource
from reprobench.utils import import_class


def bootstrap_tasks(config):
    logger.info("Bootstrapping tasks...")
    available_sources = (FileSource, UrlSource, DOISource)

    logger.trace(config)

    task_groups = {}
    for (group, task) in config["tasks"].items():
        logger.trace(f"Processing task group: {group}")

        for TaskSource in available_sources:
            if task["type"] == TaskSource.TYPE:
                source = TaskSource(**task)
                break
        else:
            raise NotImplementedError(
                f"No implementation for task source {task['type']}"
            )

        tasks = source.setup()
        task_groups[group] = [str(task) for task in tasks]

    return task_groups


def bootstrap_tools(config):
    logger.info("Setting up tools...")

    tools = {}
    for tool_name, tool in config["tools"].items():
        tools[tool_name] = dict(
            module=tool["module"], parameters=tool.get("parameters")
        )

    return tools


def bootstrap(config):
    tasks = bootstrap_tasks(config)
    tools = bootstrap_tools(config)

    return dict(tasks=tasks, tools=tools)



import atexit
import time
import sys
from multiprocessing import Pool

from sshtunnel import SSHTunnelForwarder
from loguru import logger
from tqdm import tqdm

from reprobench.core.worker import BenchmarkWorker
from reprobench.managers.base import BaseManager


class LocalManager(BaseManager):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.num_workers = kwargs.pop("num_workers")
        self.start_time = None
        self.workers = []

    def exit(self):
        for worker in self.workers:
            worker.terminate()
            worker.join()

        logger.info(f"Total time elapsed: {time.perf_counter() - self.start_time}")

    def prepare(self):
        atexit.register(self.exit)
        self.start_time = time.perf_counter()
        if self.tunneling is not None:
            self.server = SSHTunnelForwarder(
                self.tunneling["host"],
                remote_bind_address=("127.0.0.1", self.tunneling["port"]),
                ssh_pkey=self.tunneling["key_file"],
                ssh_config_file=self.tunneling["ssh_config_file"],
            )

            # https://github.com/pahaz/sshtunnel/issues/138
            if sys.version_info[0] > 3 or (
                sys.version_info[0] == 3 and sys.version_info[1] >= 7
            ):
                self.server.daemon_forward_servers = True

            self.server.start()
            self.server_address = f"tcp://127.0.0.1:{self.server.local_bind_port}"
            logger.info(f"Tunneling established at {self.server_address}")

    @staticmethod
    def spawn_worker(server_address):
        worker = BenchmarkWorker(server_address)
        worker.run()

    def spawn_workers(self):
        self.pool = Pool(self.num_workers)
        jobs = (self.server_address for _ in range(self.pending))
        self.pool_iterator = self.pool.imap_unordered(self.spawn_worker, jobs)
        self.pool.close()

    def wait(self):
        progress_bar = tqdm(desc="Executing runs", total=self.pending)
        for _ in self.pool_iterator:
            progress_bar.update()
        progress_bar.close()
        self.pool.join()

        if self.tunneling is not None:
            self.server.stop()



import click

from reprobench.console.decorators import common, server_info, use_tunneling
from reprobench.utils import read_config

from .manager import SlurmManager


@click.command("slurm")
@click.option(
    "-d", "--output-dir", type=click.Path(), default="./output", show_default=True
)
@click.option("-r", "--repeat", type=int, default=1)
@click.argument("command", type=click.Choice(("run", "stop")))
@click.argument("config", type=click.Path(), default="./benchmark.yml")
@server_info
@use_tunneling
@common
def cli(command, **kwargs):
    manager = SlurmManager(**kwargs)

    if command == "run":
        manager.run()
    elif command == "stop":
        manager.stop()


if __name__ == "__main__":
    cli()



import functools
import os
import sys
from pathlib import Path

import click
from loguru import logger


def common(func):
    @click.option("-q", "--quiet", is_flag=True)
    @click.option(
        "--verbose", "-v", "verbosity", count=True, default=0, help="Verbosity"
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        quiet = kwargs.pop("quiet")
        verbosity = kwargs.pop("verbosity")

        sys.path.append(os.getcwd())
        logger.remove()

        if not quiet:
            verbosity_levels = ["INFO", "DEBUG", "TRACE"]
            verbosity = min(verbosity, 2)
            logger.add(sys.stderr, level=verbosity_levels[verbosity])

        return func(*args, **kwargs)

    return wrapper


def server_info(func):
    @click.option("-a", "--address", default="tcp://127.0.0.1:31313", show_default=True)
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        server_address = kwargs.pop("address")
        return func(server_address=server_address, *args, **kwargs)

    return wrapper


def use_tunneling(func):
    @click.option("-h", "--host", required=False, help="[Tunneling] SSH Host")
    @click.option(
        "-p", "--port", help="[Tunneling] Remote server port for", default=31313
    )
    @click.option(
        "-K",
        "--key-file",
        type=click.Path(exists=True),
        help="[Tunneling] SSH private key file",
        required=False,
    )
    @click.option(
        "-C",
        "--ssh-config-file",
        help="[Tunneling] SSH config file",
        default=Path.home() / ".ssh" / "config",
    )
    @functools.wraps(func)
    def wrapper(host, port, key_file, ssh_config_file, *args, **kwargs):
        if host is not None:
            tunneling = dict(
                host=host, port=port, key_file=key_file, ssh_config_file=ssh_config_file
            )
            return func(tunneling=tunneling, *args, **kwargs)

        return func(tunneling=None, *args, **kwargs)

    return wrapper



"""Various utilities"""

import importlib
import re
import tarfile
import zipfile
from ast import literal_eval
from collections.abc import Iterable
from pathlib import Path
from shutil import which

import numpy
import requests
import strictyaml
from reprobench.core.exceptions import ExecutableNotFoundError, NotSupportedError
from reprobench.core.schema import schema
from retrying import retry
from tqdm import tqdm

try:
    import msgpack
    from playhouse.apsw_ext import APSWDatabase
    from reprobench.core.db import db
except ImportError:
    APSWDatabase = None
    db = None


def find_executable(executable):
    """Find an executable path from its name

    Similar to `/usr/bin/which`, this function find the path
    of an executable by its name, for example by finding it in
    the PATH environment variable.

    Args:
        executable (str): The executable name

    Returns:
        str: Path of the executable

    Raises:
        ExecutableNotFoundError: If no path for `executable`
            is found.
    """
    path = which(executable)
    if path is None:
        raise ExecutableNotFoundError
    return path


def import_class(path):
    """Import a class by its path

    Args:
        path (str): the path to the class, in similar notation as modules

    Returns:
        class: the specified class

    Examples:
        >>> import_class("reprobench.core.server.BenchmarkServer")
        <class 'reprobench.core.server.BenchmarkServer'>
    """
    module_path, tail = ".".join(path.split(".")[:-1]), path.split(".")[-1]
    module = importlib.import_module(module_path)
    return getattr(module, tail)


def _copy_file_obj(source, destination, callback, length=16 * 1024):
    """Modified version of shutil.copyfileobj with callback"""
    while True:
        buf = source.read(length)
        if not buf:
            break
        destination.write(buf)
        callback(len(buf))


def download_file(url, dest):
    """Download a file by the specified URL

    Args:
        url (str): URL for the file to download
        dest (str): Destination path for saving the file
    """
    r = requests.get(url, stream=True)

    with tqdm(
        total=int(r.headers.get("content-length", 0)),
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
    ) as progress_bar:
        progress_bar.set_postfix(file=Path(dest).name, refresh=False)
        with open(dest, "wb") as f:
            _copy_file_obj(r.raw, f, progress_bar.update)


ranged_numbers_re = re.compile(r"(?P<start>\d+)\.\.(?P<end>\d+)(\.\.(?P<step>\d+))?")


def is_range_str(range_str):
    """Check if a string is in range notation

    Args:
        range_str (str): The string to check

    Returns:
        bool: if the string is in range notation

    Examples:
        >>> is_range_str("1..2")
        True
        >>> is_range_str("1..5..2")
        True
        >>> is_range_str("1")
        False
    """
    return ranged_numbers_re.match(range_str) is not None


def str_to_range(range_str):
    """Generate range from a string with range notation

    Args:
        range_str (str): The string with range notation

    Returns:
        range: The generated range

    Examples:
        >>> str_to_range("1..3")
        range(1, 4)
        >>> str_to_range("1..5..2")
        range(1, 6, 2)
        >>> [*str_to_range("1..3")]
        [1, 2, 3]
    """
    matches = ranged_numbers_re.match(range_str).groupdict()
    start = int(matches["start"])
    end = int(matches["end"]) + 1

    if matches["step"]:
        return range(start, end, int(matches["step"]))
    return range(start, end)


def encode_message(obj):
    """Encode an object for transport

    This method serialize the object with msgpack for
    network transportation.

    Args:
        obj: serializable object

    Returns:
        bin: binary string of the encoded object
    """
    return msgpack.packb(obj, use_bin_type=True)


def decode_message(msg):
    """Decode an encoded object

    This method deserialize the encoded object from
    `encode_message(obj)`.

    Args:
        bin: binary string of the encoded object

    Returns:
        obj: decoded object
    """
    return msgpack.unpackb(msg, raw=False)


@retry(wait_exponential_multiplier=500)
def send_event(socket, event_type, payload=None):
    """Used in the worker with a DEALER socket to send events to the server.

    Args:
        socket (zmq.Socket): the socket for sending the event
        event_type (str): event type agreed between the parties
        payload (any, optional): the payload for the event
    """
    event = [event_type, encode_message(payload)]
    socket.send_multipart(event)


def recv_event(socket):
    """Receive published event for the observers

    Args:
        socket (zmq.Socket): SUB socket for receiving the event

    Returns:
        (event_type, payload, address): Tuple for received events
    """
    event_type, payload, address = socket.recv_multipart()

    return event_type, decode_message(payload), address


def get_db_path(output_dir):
    """Get the database path from the given output directory

    Args:
        output_dir (str): path to the output directory

    Returns:
        str: database path
    """
    return str((Path(output_dir) / f"benchmark.db").resolve())


def init_db(db_path):
    """Initialize the given database

    Args:
        db_path (str): path to the database
    """
    database = APSWDatabase(db_path, pragmas=(("journal_mode", "wal"),))
    db.initialize(database)


def resolve_files_uri(root):
    """Resolve all `file://` URIs in a dictionary to its content

    Args:
        root (dict): Root dictionary of the configuration

    Examples:
        >>> d = dict(test="file://./test.txt")
        >>> resolve_files_uri(d)
        >>> d
        {'a': 'this is the content of test.txt\\n'}
    """
    protocol = "file://"
    iterator = None
    if isinstance(root, dict):
        iterator = root
    elif isinstance(root, list) or isinstance(root, tuple):
        iterator = range(len(root))

    for k in iterator:
        if isinstance(root[k], str) and root[k].startswith(protocol):
            root[k] = Path(root[k][len(protocol) :]).read_text()
        elif isinstance(root[k], Iterable) and not isinstance(root[k], str):
            resolve_files_uri(root[k])


def read_config(config_path, resolve_files=False):
    """Read a YAML configuration from a path

    Args:
        config_path (str): Configuration file path (YAML)
        resolve_files (bool, optional): Should files be resolved to its content? Defaults to False.

    Returns:
        dict: Configuration
    """
    with open(config_path, "r") as f:
        config_text = f.read()
        config = strictyaml.load(config_text, schema=schema).data

    if resolve_files:
        resolve_files_uri(config)

    return config


def extract_zip(path, dest):
    """Extract a ZIP file

    Args:
        path (str): Path to ZIP file
        dest (str): Destination for extraction
    """
    if not dest.is_dir():
        with zipfile.ZipFile(path, "r") as f:
            f.extractall(dest)


def extract_tar(path, dest):
    """Extract a TAR file

    Args:
        path (str): Path to TAR file
        dest (str): Destination for extraction
    """
    if not dest.is_dir():
        with tarfile.TarFile.open(path) as f:
            f.extractall(dest)


def extract_archives(path):
    """Extract archives based on its extension

    Args:
        path (str): Path to the archive file
    """
    extract_path = Path(path).with_name(path.stem)

    if zipfile.is_zipfile(path):
        extract_zip(path, extract_path)
    elif tarfile.is_tarfile(path):
        extract_tar(path, extract_path)


def get_pcs_parameter_range(parameter_str, is_categorical):
    """Generate a range from specified pcs range notation

    Args:
        parameter_str (str): specified pcs parameter
        is_categorical (bool): is the range categorical

    Raises:
        NotSupportedError: If there is no function for resolving the range

    Returns:
        range: Generated range
    """
    functions = dict(
        range=range,
        arange=numpy.arange,
        linspace=numpy.linspace,
        logspace=numpy.logspace,
        geomspace=numpy.geomspace,
    )

    function_re = re.compile(r"(?P<function>[A-Za-z_]+)\((?P<arguments>.*)\)")

    match = function_re.match(parameter_str)

    parameter_range = None
    if match:
        function = match.group("function")
        if function not in functions:
            raise NotSupportedError(f"Declaring range with {function} is not supported")
        args = literal_eval(match.group("arguments"))
        parameter_range = functions[function](*args)
    else:
        parameter_range = literal_eval(parameter_str)
        if not isinstance(parameter_range, Iterable) or isinstance(
            parameter_range, str
        ):
            parameter_range = (parameter_range,)
        if is_categorical:
            parameter_range = map(str, parameter_range)

    return parameter_range


def parse_pcs_parameters(lines):
    """Parse parameters from a pcs file content

    Args:
        lines ([str]): pcs file content

    Returns:
        dict: generated parameters
    """
    parameter_range_indicator = "-->"

    parameters = {}
    parameter_key = None
    is_categorical = False

    for line in lines:
        if ("{" in line or "[" in line) and not line.startswith("#"):
            parameter_key = line[: line.find(" ")]
            is_categorical = "{" in line

        if "#" not in line or parameter_range_indicator not in line:
            continue

        comment_pos = line.find("#")
        pos = line.find(parameter_range_indicator, comment_pos)
        parameter_str = line[pos + len(parameter_range_indicator) :].strip()

        parameter_range = get_pcs_parameter_range(parameter_str, is_categorical)

        parameters[parameter_key] = parameter_range

    return parameters


def check_valid_config_space(config_space, parameters):
    """Check if the parameters is valid based on a configuration space

    Args:
        config_space (ConfigSpace): configuration space
        parameters (dict): parameters dictionary

    Raises:
        ValueError: If there is invalid values
    """
    base = config_space.get_default_configuration()
    for key, value in parameters.items():
        if key in base:
            base[key] = value




import zmq
from loguru import logger

from reprobench.core.events import BOOTSTRAP
from reprobench.core.bootstrap import bootstrap_client
from reprobench.utils import decode_message, read_config, send_event


class BaseManager(object):
    def __init__(self, config, server_address, tunneling, **kwargs):
        self.server_address = server_address
        self.tunneling = tunneling
        self.config = read_config(config, resolve_files=True)
        self.output_dir = kwargs.pop("output_dir")
        self.repeat = kwargs.pop("repeat")

        context = zmq.Context()
        self.socket = context.socket(zmq.DEALER)

    def prepare(self):
        pass

    def spawn_workers(self):
        raise NotImplementedError

    def bootstrap(self):
        self.socket.connect(self.server_address)

        client_results = bootstrap_client(self.config)
        bootstrapped_config = {**self.config, **client_results}

        logger.info("Sending bootstrap event to server")
        payload = dict(
            config=bootstrapped_config, output_dir=self.output_dir, repeat=self.repeat
        )
        send_event(self.socket, BOOTSTRAP, payload)

        self.pending = decode_message(self.socket.recv())

    def wait(self):
        pass

    def stop(self):
        pass

    def run(self):
        self.prepare()
        self.bootstrap()
        self.spawn_workers()
        self.wait()



import click
from loguru import logger
from playhouse.apsw_ext import APSWDatabase

from reprobench.console.decorators import common
from reprobench.core.db import Step, db
from reprobench.utils import get_db_path, import_class, init_db, read_config


class BenchmarkAnalyzer(object):
    def __init__(self, output_dir, config, **kwargs):
        self.output_dir = output_dir
        self.config = read_config(config)
        self.db_path = get_db_path(output_dir)
        init_db(self.db_path)

    def run(self):
        steps = self.config["steps"]["analysis"]
        context = dict(output_dir=self.output_dir, db_path=self.db_path)
        for step in steps:
            logger.debug(f"Running {step['module']}")
            module = import_class(step["module"])
            module.execute(context, step["config"])


@click.command(name="analyze")
@click.option(
    "-d", "--output-dir", type=click.Path(), default="./output", show_default=True
)
@click.argument("config", type=click.Path(), default="./benchmark.yml")
@common
def cli(**kwargs):
    analyzer = BenchmarkAnalyzer(**kwargs)
    analyzer.run()


if __name__ == "__main__":
    cli()
